!MNH_LIC Copyright 2011-2022 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!     #####################
      MODULE MODI_GRAVITY_IMPL
!     #####################
!
INTERFACE
      SUBROUTINE GRAVITY_IMPL (HLBCX, HLBCY,KRR,KRRL,KRRI, PTSTEP,             &
                            PTHT, PRT, PTHVREF, PRHODJ,                        &
                            PRWS, PRTHS, PRRS, PRTHS_CLD, PRRS_CLD             )
!
CHARACTER(LEN=4),DIMENSION(2),INTENT(IN):: HLBCX, HLBCY  ! X- and Y-direc LBC
!
INTEGER,                  INTENT(IN)    :: KRR  ! Number of moist variables
INTEGER,                  INTENT(IN)    :: KRRL ! Number of liquid water var.
INTEGER,                  INTENT(IN)    :: KRRI ! Number of ice water var.
!
REAL,                     INTENT(IN)    :: PTSTEP
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT, PRHODJ
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRT 
                                                  ! Variables at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHVREF   ! Virtual Temperature
                                          ! of the reference state
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRWS
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRTHS
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRRS 
                                          ! Sources terms 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRTHS_CLD
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRRS_CLD
                                          ! tendencies from previous time-step
                                          ! cloud processes
!
END SUBROUTINE GRAVITY_IMPL
!
END INTERFACE
!
END MODULE MODI_GRAVITY_IMPL 
!     ##########################################################################
      SUBROUTINE GRAVITY_IMPL (HLBCX, HLBCY,KRR,KRRL,KRRI, PTSTEP,             &
                            PTHT, PRT, PTHVREF, PRHODJ,                        &
                            PRWS, PRTHS, PRRS, PRTHS_CLD, PRRS_CLD             )
!     ##########################################################################
!
!!****  *GRAVITY_IMPL * - routine to estimate gravity term using future buoyancy
!!
!!    PURPOSE
!!    -------
!!
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!      NONE
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!	V. Masson        * Meteo France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original    04/2011 
!!      Q.Rodier 06/15 correction on budget
!  P. Wautelet    02/2020: use the new data structures and subroutines for budgets
!
!-------------------------------------------------------------------------------
!
!*       0.    DECLARATIONS
!              ------------
!
use modd_budget,         only: lbudget_w, NBUDGET_W, tbudgets

use mode_budget,         only: Budget_store_init, Budget_store_end
use mode_mppdb
#ifdef MNH_OPENACC
use mode_mnh_zwork,      only: Mnh_mem_get, Mnh_mem_position_pin, Mnh_mem_release
#endif

use modi_adv_boundaries
use modi_gravity

!-------------------------------------------------------------------------------
!
IMPLICIT NONE
!
!*       0.1   Declarations of dummy arguments :
!
CHARACTER(LEN=4),DIMENSION(2),INTENT(IN):: HLBCX, HLBCY  ! X- and Y-direc LBC
!
INTEGER,                  INTENT(IN)    :: KRR  ! Number of moist variables
INTEGER,                  INTENT(IN)    :: KRRL ! Number of liquid water var.
INTEGER,                  INTENT(IN)    :: KRRI ! Number of ice water var.
!
REAL,                     INTENT(IN)    :: PTSTEP
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHT, PRHODJ
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRT
                                                  ! Variables at t
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PTHVREF   ! Virtual Temperature
                                          ! of the reference state
REAL, DIMENSION(:,:,:),   INTENT(INOUT) :: PRWS
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRTHS
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRRS 
                                          ! Sources terms 
REAL, DIMENSION(:,:,:),   INTENT(IN)    :: PRTHS_CLD
REAL, DIMENSION(:,:,:,:), INTENT(IN)    :: PRRS_CLD
                                          ! tendencies from previous time-step
                                          ! cloud processes
!
!
!*       0.2   declarations of local variables
!
!  
! Tendencies of W due to gravity
REAL, DIMENSION(:,:,:),   POINTER, CONTIGUOUS :: ZRWS_GRAV
! Guess of future theta
REAL, DIMENSION(:,:,:),   POINTER, CONTIGUOUS :: ZTH
! Guess of future mixing ratios
REAL, DIMENSION(:,:,:,:), POINTER, CONTIGUOUS :: ZR
!
INTEGER :: JR
!
INTEGER  :: JI,JJ,JK
INTEGER  :: JIU, JJU, JKU, JRU
!
!-------------------------------------------------------------------------------

IF (MPPDB_INITIALIZED) THEN
  !Check all IN arrays
  CALL MPPDB_CHECK(PTHT,     "GRAVITY_IMPL beg:PTHT")
  CALL MPPDB_CHECK(PRHODJ,   "GRAVITY_IMPL beg:PRHODJ")
  CALL MPPDB_CHECK(PRT,      "GRAVITY_IMPL beg:PRT")
  CALL MPPDB_CHECK(PTHVREF,  "GRAVITY_IMPL beg:PTHVREF")
  CALL MPPDB_CHECK(PRTHS,    "GRAVITY_IMPL beg:PRTHS")
  CALL MPPDB_CHECK(PRRS,     "GRAVITY_IMPL beg:PRRS")
  CALL MPPDB_CHECK(PRTHS_CLD,"GRAVITY_IMPL beg:PRTHS_CLD")
  CALL MPPDB_CHECK(PRRS_CLD, "GRAVITY_IMPL beg:PRRS_CLD")
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRWS,     "GRAVITY_IMPL beg:PRWS")
END IF

JIU =  size(ptht, 1 )
JJU =  size(ptht, 2 )
JKU =  size(ptht, 3 )
JRU =  size( prt, 4 )

#ifndef MNH_OPENACC
allocate( zrws_grav(JIU, JJU, JKU ) )
allocate( zth      (JIU, JJU, JKU ) )
allocate( zr       (JIU, JJU, JKU, JRU ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN()

CALL MNH_MEM_GET( zrws_grav, JIU, JJU, JKU )
CALL MNH_MEM_GET( zth,       JIU, JJU, JKU )
CALL MNH_MEM_GET( zr,        JIU, JJU, JKU, JRU )
#endif

!$acc data present( PTHT, PRHODJ, PRT, PTHVREF, PRWS, PRTHS, PRRS, PRTHS_CLD, PRRS_CLD, &
!$acc &             zrws_grav, zth, zr )

if ( lbudget_w ) call Budget_store_init( tbudgets(NBUDGET_W), 'GRAV', prws(:, :, :) )

! guess of Theta at future time-step
!$acc kernels
ZTH(:,:,:) = (PRTHS(:,:,:) + PRTHS_CLD(:,:,:)) / PRHODJ(:,:,:) * PTSTEP
!
DO CONCURRENT (  JI=1:JIU,JJ=1:JJU,JK=1:JKU , JR = 1:KRR )
 ZR(JI,JJ,JK,JR) = (PRRS(JI,JJ,JK,JR) + PRRS_CLD(JI,JJ,JK,JR)) / PRHODJ(JI,JJ,JK) * PTSTEP
END DO ! CONCURRENT
!$acc end kernels
!
#ifndef MNH_OPENACC
CALL ADV_BOUNDARIES( HLBCX, HLBCY, ZTH, PTHT )
DO JR = 1, KRR
  CALL ADV_BOUNDARIES( HLBCX, HLBCY, ZR(:,:,:,JR), PRT(:,:,:,JR) )
END DO
#else
CALL ADV_BOUNDARIES_DEVICE( HLBCX, HLBCY, ZTH, PTHT )
DO JR = 1, KRR
  CALL ADV_BOUNDARIES_DEVICE(HLBCX, HLBCY, ZR(:,:,:,JR), PRT(:,:,:,JR) )
END DO
#endif
!
! ====> A vérifier si c'est nécessaire d'échanger les champs
!       A priori, je dirai non.
!
! gravity effect on vertical speed
CALL GRAVITY ( KRR,KRRL, KRRI, ZTH, ZR, PRHODJ, PTHVREF, ZRWS_GRAV(:,:,:) )
!
!$acc kernels
DO CONCURRENT ( JI=1:JIU,JJ=1:JJU,JK=1:JKU )
   PRWS(JI,JJ,JK) = PRWS(JI,JJ,JK) + ZRWS_GRAV(JI,JJ,JK)
END DO
!$acc end kernels
!
if ( lbudget_w ) then
!$acc update self(PRWS)
  call Budget_store_end( tbudgets(NBUDGET_W), 'GRAV', prws(:, :, :) )
end if

!$acc end data

#ifndef MNH_OPENACC
DEALLOCATE( zrws_grav, zth, zr )
#else
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE()
#endif

IF (MPPDB_INITIALIZED) THEN
  !Check all INOUT arrays
  CALL MPPDB_CHECK(PRWS,"GRAVITY_IMPL end:PRWS")
END IF

!-------------------------------------------------------------------------------
!
END SUBROUTINE GRAVITY_IMPL
