!MNH_LIC Copyright 1995-2019 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!    #######################  
     MODULE MODI_ROTATE_WIND
!    #######################
!
INTERFACE 
!
      SUBROUTINE ROTATE_WIND(PU,PV,PW,                          &
                             PDIRCOSXW, PDIRCOSYW, PDIRCOSZW,   &
                             PCOSSLOPE,PSINSLOPE,               &
                             PDXX,PDYY,PDZZ,                    &
                             PUSLOPE,PVSLOPE                    )
!
!
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PU,PV,PW        ! cartesian components
                                 ! of the wind
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSXW, PDIRCOSYW, PDIRCOSZW
! Director Cosinus along x, y and z directions at surface w-point
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE       ! cosinus of the angle 
                                 ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE       ! sinus of the angle 
                                 ! between i and the slope vector
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDXX, PDYY, PDZZ
                                 ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(OUT)  ::  PUSLOPE         ! wind component along 
                                 ! the maximum slope direction
REAL, DIMENSION(:,:),   INTENT(OUT)  ::  PVSLOPE         ! wind component along
                                 !  the direction normal to the maximum slope one
!
!-------------------------------------------------------------------------------
!
END SUBROUTINE ROTATE_WIND
!
END INTERFACE
!
END MODULE MODI_ROTATE_WIND
!
!     ###########################################################
      SUBROUTINE ROTATE_WIND(PU,PV,PW,                          &
                             PDIRCOSXW, PDIRCOSYW, PDIRCOSZW,   &
                             PCOSSLOPE,PSINSLOPE,               &
                             PDXX,PDYY,PDZZ,                    &
                             PUSLOPE,PVSLOPE                    )
!     ###########################################################
!
!
!!****  *ROTATE_WIND* - computes the wind components along the maximum slope 
!!               direction and its normal direction in the first mass level.
!!
!!    PURPOSE
!!    -------
!!**** 
!        The purpose of this routine is to compute the wind component parallel 
!     to the orography at the first mass level. The exact location where these
!     components are computed is the point of intersection between the normal 
!     to the orography and the first mass-level hyper-plane at PDZZ(:,:,IKB)/2 
!        
!!**  METHOD
!!    ------
!!       The values of the 3 cartesian components of the wind are determined
!!    by a bilinear interpolation between the 4 nearest points in the first 
!!    mass-level hyper-plane. These points are found according to the signs of 
!!    the slopes' sinus and cosinus. For each direction of interpolation, the 
!!    two different localizations (mass or flux grids) are used to avoid 
!!    lateral boundary problems.  
!!       Then, the rotation is performed for the wind components. The rotation 
!!    angle is the angle between the x axe and the maximum slope direction 
!!    defined by the slope vector (dZs/dx , dZs/dy).
!!        Finally, the horizontal components are set at the marginal points 
!!    according to cyclic boundary conditions because this is the only case
!!    where these points can be considered.
!!
!!    EXTERNAL
!!    --------
!!       NONE
!!
!!    IMPLICIT ARGUMENTS 
!!    ------------------
!!
!!       MODD_CONF      : L2D   switch for 2D model version
!!
!!
!!    REFERENCE
!!    ---------
!!      Book 1 of documentation (Chapter: Turbulence)
!!
!!    AUTHOR
!!    ------
!!      Joel Stein              * Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!      Original         14/11/95
!!      Modifications:   15/05/96, (N. wood)
!!                                 take into account no slip conditions 
!!                                 at the surface
!!                       14/02/01  (V. Masson)
!!                                 Slip condition at the surface restored
!!
!! --------------------------------------------------------------------------
!       
!*      0. DECLARATIONS
!          ------------
USE MODD_PARAMETERS

#ifdef MNH_OPENACC
USE MODE_MNH_ZWORK, ONLY: MNH_MEM_GET, MNH_MEM_POSITION_PIN, MNH_MEM_RELEASE
#endif
use mode_mppdb

#if defined(MNH_BITREP) || defined(MNH_BITREP_OMP)
USE MODI_BITREP
#endif
!
IMPLICIT NONE
!
!
!*      0.1  declarations of arguments
!
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PU,PV,PW        ! cartesian components
                                 ! of the wind
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PDIRCOSXW, PDIRCOSYW, PDIRCOSZW
! Director Cosinus along x, y and z directions at surface w-point
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PCOSSLOPE       ! cosinus of the angle 
                                 ! between i and the slope vector
REAL, DIMENSION(:,:),   INTENT(IN)   ::  PSINSLOPE       ! sinus of the angle 
                                 ! between i and the slope vector
REAL, DIMENSION(:,:,:), INTENT(IN)   ::  PDXX, PDYY, PDZZ
                                 ! Metric coefficients
REAL, DIMENSION(:,:),   INTENT(OUT)  ::  PUSLOPE         ! wind component along 
                                 ! the maximum slope direction
REAL, DIMENSION(:,:),   INTENT(OUT)  ::  PVSLOPE         ! wind component along
                                 !  the direction normal to the maximum slope one
!
!-------------------------------------------------------------------------------
!
!       0.2  declaration of local variables
!
#ifndef MNH_OPENACC
INTEGER, DIMENSION(:,:), allocatable :: ILOC,JLOC             ! shift index to find the 4 nearest points in x and y directions
REAL,    DIMENSION(:,:), allocatable :: ZCOEFF,ZCOEFM,     &  ! interpolation weigths for flux and mass locations
                                        ZUINT,ZVINT,ZWINT, &  ! intermediate values of the cartesian components after x interp.
                                        ZUFIN,ZVFIN,ZWFIN, &  ! final values of the cartesian components after the 2 interp.
                                        ZWGROUND              ! vertical velocity at the surface
#else
INTEGER, DIMENSION(:,:), pointer, contiguous :: ILOC,JLOC             ! shift index to find the 4 nearest points in x and y directions
REAL,    DIMENSION(:,:), pointer, contiguous :: ZCOEFF,ZCOEFM,     &  ! interpolation weigths for flux and mass locations
                                                ZUINT,ZVINT,ZWINT, &  ! intermediate values of the cartesian components after x interp.
                                                ZUFIN,ZVFIN,ZWFIN, &  ! final values of the cartesian components after the 2 interp.
                                                ZWGROUND              ! vertical velocity at the surface
#endif
INTEGER     :: IIB,IIE,IJB,IJE,IKB ! index values for the Beginning or the End of the physical domain in x,y and z directions
INTEGER     :: IIU,IJU             ! arrays' sizes for i and j indices
INTEGER     :: JI,JJ
!
!----------------------------------------------------------------------------

if ( mppdb_initialized ) then
  !check all in arrays
  call Mppdb_check( pu,        "Rotate_wind beg:pu"        )
  call Mppdb_check( pv,        "Rotate_wind beg:pv"        )
  call Mppdb_check( pw,        "Rotate_wind beg:pw"        )
  call Mppdb_check( pdircosxw, "Rotate_wind beg:pdircosxw" )
  call Mppdb_check( pdircosyw, "Rotate_wind beg:pdircosyw" )
  call Mppdb_check( pdircoszw, "Rotate_wind beg:pdircoszw" )
  call Mppdb_check( pcosslope, "Rotate_wind beg:pcosslope" )
  call Mppdb_check( psinslope, "Rotate_wind beg:psinslope" )
  call Mppdb_check( pdxx,      "Rotate_wind beg:pdxx"      )
  call Mppdb_check( pdyy,      "Rotate_wind beg:pdyy"      )
  call Mppdb_check( pdzz,      "Rotate_wind beg:pdzz"      )
end if

!$acc data present( PU, PV, PW, PDIRCOSXW, PDIRCOSYW, PDIRCOSZW,  &
!$acc &             PCOSSLOPE, PSINSLOPE, PDXX, PDYY, PDZZ, PUSLOPE, PVSLOPE )

#ifndef MNH_OPENACC
allocate( iloc    (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( jloc    (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zcoeff  (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zcoefm  (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zuint   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zvint   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zwint   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zufin   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zvfin   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zwfin   (size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
allocate( zwground(size( pdircosxw, 1 ), size( pdircosxw, 2 ) ) )
#else
!Pin positions in the pools of MNH memory
CALL MNH_MEM_POSITION_PIN( 'ROTATE_WIND' )

CALL MNH_MEM_GET( iloc,     size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( jloc,     size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zcoeff,   size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zcoefm,   size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zuint,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zvint,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zwint,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zufin,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zvfin,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zwfin,    size( pdircosxw, 1 ), size( pdircosxw, 2 ) )
CALL MNH_MEM_GET( zwground, size( pdircosxw, 1 ), size( pdircosxw, 2 ) )

!$acc data present( iloc, jloc, zcoeff, zcoefm, zuint, zvint, zwint, zufin, zvfin, zwfin, zwground )
#endif

!
!*      1.    PRELIMINARIES
!             -------------
!
!$acc kernels
PUSLOPE=0.
PVSLOPE=0.
!
IIB = 2
IJB = 2
IIU = SIZE(PU,1)
IJU = SIZE(PU,2)
IIE = IIU - 1
IJE = IJU - 1
IKB = 1+JPVEXT
!
ZWGROUND(:,:) = PW(:,:,IKB)
!
!*      2.    INTERPOLATE THE CARTESIAN COMPONENTS
!             ------------------------------------
!
ILOC(:,:)=NINT(SIGN(1.,-PCOSSLOPE(:,:)))
JLOC(:,:)=NINT(SIGN(1.,-PSINSLOPE(:,:)))
!
! interpolation in x direction
!
!$acc_nv loop independent collapse(2)
DO CONCURRENT(JI=IIB:IIE,JJ=1:IJU)
    ZCOEFF(JI,JJ) =                                                  &
      (0.5*PDXX(JI,JJ,IKB) + 0.5*PDZZ(JI,JJ,IKB)*PDIRCOSXW(JI,JJ) )  & 
      * 2. / (PDXX(JI,JJ,IKB)+PDXX(JI+1,JJ,IKB))
    ZUINT(JI,JJ) = ZCOEFF(JI,JJ)      * PU(JI+1,JJ,IKB)  +           &
                   (1.-ZCOEFF(JI,JJ)) * PU(JI,JJ,IKB)
    !
    ZCOEFM(JI,JJ) = 1. - 0.5 * PDZZ(JI,JJ,IKB) * ABS(PDIRCOSXW(JI,JJ))      & 
                             / PDXX(JI+(ILOC(JI,JJ)+1)/2,JJ,IKB)
    ZVINT(JI,JJ) = ZCOEFM(JI,JJ)      * PV(JI,JJ,IKB)              +        &
                   (1.-ZCOEFM(JI,JJ)) * PV(JI+ILOC(JI,JJ),JJ,IKB)
    !
    ZWINT(JI,JJ) = ZCOEFM(JI,JJ)  * (PW(JI,JJ,IKB+1)+ZWGROUND(JI,JJ)) * 0.5    &
              + (1.-ZCOEFM(JI,JJ))                                             &
               *(PW(JI+ILOC(JI,JJ),JJ,IKB+1)+ZWGROUND(JI+ILOC(JI,JJ),JJ)) * 0.5
END DO
!
! interpolation in y direction
!
!$acc_nv loop independent collapse(2)
DO CONCURRENT(JI=IIB:IIE,JJ=IJB:IJE)
    ZCOEFF(JI,JJ) =                                                     &
      (0.5*PDYY(JI,JJ,IKB) + 0.5*PDZZ(JI,JJ,IKB)*PDIRCOSYW(JI,JJ) )     & 
      * 2. / (PDYY(JI,JJ,IKB)+PDYY(JI+1,JJ,IKB))
    ZVFIN(JI,JJ) = ZCOEFF(JI,JJ)      * ZVINT(JI,JJ+1)  +               &
                   (1.-ZCOEFF(JI,JJ)) * ZVINT(JI,JJ)
    !
    ZCOEFM(JI,JJ) = 1. - 0.5 * PDZZ(JI,JJ,IKB) * ABS(PDIRCOSYW(JI,JJ))   & 
                             / PDYY(JI,JJ+(JLOC(JI,JJ)+1)/2,IKB)
    ZUFIN(JI,JJ) = ZCOEFM(JI,JJ)      * ZUINT(JI,JJ)                +    &
                   (1.-ZCOEFM(JI,JJ)) * ZUINT(JI,JJ+JLOC(JI,JJ))
    ZWFIN(JI,JJ) = ZCOEFM(JI,JJ)      * ZWINT(JI,JJ)                +    &
                   (1.-ZCOEFM(JI,JJ)) * ZWINT(JI,JJ+JLOC(JI,JJ))
END DO
!
!*      3.    ROTATE THE WIND
!             ---------------
!
!
!$acc_nv loop independent collapse(2)
DO CONCURRENT(JI=IIB:IIE,JJ=IJB:IJE)
    PUSLOPE(JI,JJ) = PCOSSLOPE(JI,JJ) * PDIRCOSZW(JI,JJ) * ZUFIN(JI,JJ) +   &
                     PSINSLOPE(JI,JJ) * PDIRCOSZW(JI,JJ) * ZVFIN(JI,JJ) +   &
#if !defined(MNH_BITREP) && !defined(MNH_BITREP_OMP)
                            SQRT(1.-PDIRCOSZW(JI,JJ)**2) * ZWFIN(JI,JJ)
#else
                            SQRT(1.-BR_P2(PDIRCOSZW(JI,JJ))) * ZWFIN(JI,JJ)
#endif
    !              
    PVSLOPE(JI,JJ) =-PSINSLOPE(JI,JJ)                    * ZUFIN(JI,JJ) +   &
                     PCOSSLOPE(JI,JJ)                    * ZVFIN(JI,JJ)
    !
END DO
!$acc end kernels

!$acc end data

#ifdef MNH_OPENACC
!Release all memory allocated with MNH_MEM_GET calls since last call to MNH_MEM_POSITION_PIN
CALL MNH_MEM_RELEASE( 'ROTATE_WIND' )
#endif

!$acc end data

if ( mppdb_initialized ) then
  !check all out arrays
  call Mppdb_check( puslope, "Rotate_wind end:puslope" )
  call Mppdb_check( pvslope, "Rotate_wind end:pvslope" )
end if

!----------------------------------------------------------------------------
!
END SUBROUTINE ROTATE_WIND
